(* base 2^30 *)
(* all mod p where p = 115792089237316195423570985008687907853269984665640564039457584007908834671663 *)
type intp = int * int * int * int * int * int * int * int * int

let zero_intp = (0,0,0,0,0,0,0,0,0)
let unit_intp = (0,0,0,0,0,0,0,0,1)
let two_intp = (0,0,0,0,0,0,0,0,2)
let p = (65535,1073741823,1073741823,1073741823,1073741823,1073741823,1073741823,1073741819,1073740847)
let pa = [| 65535;1073741823;1073741823;1073741823;1073741823;1073741823;1073741823;1073741819;1073740847 |]
let pminus2_intp = (65535,1073741823,1073741823,1073741823,1073741823,1073741823,1073741823,1073741819,1073740845)

let rega : int array = Array.make 9 0
let regb : int array = Array.make 9 0
let regc : int array = Array.make 9 0
let regd : int array = Array.make 9 0
let rege : int array = Array.make 9 0
let regf : int array = Array.make 9 0
let regg : int array = Array.make 9 0
let regh : int array = Array.make 9 0

let lt_reg_p reg =
  if reg.(0) < 65535 then
    true
  else if reg.(0) > 65535 then
    false
  else
    if reg.(1) < 1073741823 then
      true
    else if reg.(1) > 1073741823 then
      false
    else
      if reg.(2) < 1073741823 then
	true
      else if reg.(2) > 1073741823 then
	false
      else
	if reg.(3) < 1073741823 then
	  true
	else if reg.(3) > 1073741823 then
	  false
	else
	  if reg.(4) < 1073741823 then
	    true
	  else if reg.(4) > 1073741823 then
	    false
	  else
	    if reg.(5) < 1073741823 then
	      true
	    else if reg.(5) > 1073741823 then
	      false
	    else
	      if reg.(6) < 1073741823 then
		true
	      else if reg.(6) > 1073741823 then
		false
	      else
		if reg.(7) < 1073741819 then
		  true
		else if reg.(7) > 1073741819 then
		  false
		else
		  reg.(8) < 1073740847

let sub_reg_p reg =
  let b = ref false in
  for i = 8 downto 0 do
    if !b then
      reg.(i) <- (reg.(i) - pa.(i)) - 1
    else
      reg.(i) <- reg.(i) - pa.(i);
    if reg.(i) < 0 then
      begin
	reg.(i) <- reg.(i) land 0x3fffffff;
	b := true
      end
    else
      b := false
  done

(** assume reg is in [0,2p) and change to be in [0,p) **)
let ensure_reg_modp reg = if not (lt_reg_p reg) then sub_reg_p reg

let reg_to_intp reg = (reg.(0),reg.(1),reg.(2),reg.(3),reg.(4),reg.(5),reg.(6),reg.(7),reg.(8))
let set_reg_intp reg (x8,x7,x6,x5,x4,x3,x2,x1,x0) =
  reg.(0) <- x8;
  reg.(1) <- x7;
  reg.(2) <- x6;
  reg.(3) <- x5;
  reg.(4) <- x4;
  reg.(5) <- x3;
  reg.(6) <- x2;
  reg.(7) <- x1;
  reg.(8) <- x0

let copy_reg_reg_intp reg1 reg2 =
  reg2.(0) <- reg1.(0);
  reg2.(1) <- reg1.(1);
  reg2.(2) <- reg1.(2);
  reg2.(3) <- reg1.(3);
  reg2.(4) <- reg1.(4);
  reg2.(5) <- reg1.(5);
  reg2.(6) <- reg1.(6);
  reg2.(7) <- reg1.(7);
  reg2.(8) <- reg1.(8)

let intp_i_r (x8,x7,x6,x5,x4,x3,x2,x1,x0) i =
  if i = 0 then x8
  else if i = 1 then x7
  else if i = 2 then x6
  else if i = 3 then x5
  else if i = 4 then x4
  else if i = 5 then x3
  else if i = 6 then x2
  else if i = 7 then x1
  else x0
      
(* add x to register with result in reg *)
let add_reg_intp reg x =
  let c = ref 0 in
  for i = 8 downto 0 do
    reg.(i) <- reg.(i) + intp_i_r x i + !c;
    if reg.(i) land 0x40000000 = 0 then
      c := 0
    else
      begin
	reg.(i) <- reg.(i) land 0x3fffffff;
	c := 1
      end
  done;
  ensure_reg_modp reg

(* add two registers with result in reg2 *)
let add_reg_reg_intp reg1 reg2 =
  let c = ref 0 in
  for i = 8 downto 0 do
    reg2.(i) <- reg2.(i) + reg1.(i) + !c;
    if reg2.(i) land 0x40000000 = 0 then
      c := 0
    else
      begin
	reg2.(i) <- reg2.(i) land 0x3fffffff;
	c := 1
      end
  done;
  ensure_reg_modp reg2

(* set reg to p - reg *)
let minus_reg_intp reg =
  reg.(0) <- pa.(0) - reg.(0);
  for i = 1 to 8 do
    reg.(i) <- pa.(i) - reg.(i);
    if reg.(i) < 0 then
      begin
	reg.(i-1) <- reg.(i-1) - 1;
	reg.(i) <- reg.(i) land 0x3FFFFFFF
      end
  done

let reg_bit reg i =
  if i < 30 then
    (reg.(8) lsr i) land 1 = 1
  else if i < 60 then
    (reg.(7) lsr (i-30)) land 1 = 1
  else if i < 90 then
    (reg.(6) lsr (i-60)) land 1 = 1
  else if i < 120 then
    (reg.(5) lsr (i-90)) land 1 = 1
  else if i < 150 then
    (reg.(4) lsr (i-120)) land 1 = 1
  else if i < 180 then
    (reg.(3) lsr (i-150)) land 1 = 1
  else if i < 210 then
    (reg.(2) lsr (i-180)) land 1 = 1
  else if i < 240 then
    (reg.(1) lsr (i-210)) land 1 = 1
  else
    (reg.(0) lsr (i-240)) land 1 = 1

let intp_bit (x8,x7,x6,x5,x4,x3,x2,x1,x0) i =
  if i < 30 then
    (x0 lsr i) land 1 = 1
  else if i < 60 then
    (x1 lsr (i-30)) land 1 = 1
  else if i < 90 then
    (x2 lsr (i-60)) land 1 = 1
  else if i < 120 then
    (x3 lsr (i-90)) land 1 = 1
  else if i < 150 then
    (x4 lsr (i-120)) land 1 = 1
  else if i < 180 then
    (x5 lsr (i-150)) land 1 = 1
  else if i < 210 then
    (x6 lsr (i-180)) land 1 = 1
  else if i < 240 then
    (x7 lsr (i-210)) land 1 = 1
  else
    (x8 lsr (i-240)) land 1 = 1

let boolarray : bool array = Array.make 256 false
let intpb : intp array =
  [|
    (0,0,0,0,0,0,0,0,1);
    (0,0,0,0,0,0,0,0,2);
    (0,0,0,0,0,0,0,0,4);
    (0,0,0,0,0,0,0,0,8);
    (0,0,0,0,0,0,0,0,16);
    (0,0,0,0,0,0,0,0,32);
    (0,0,0,0,0,0,0,0,64);
    (0,0,0,0,0,0,0,0,128);
    (0,0,0,0,0,0,0,0,256);
    (0,0,0,0,0,0,0,0,512);
    (0,0,0,0,0,0,0,0,1024);
    (0,0,0,0,0,0,0,0,2048);
    (0,0,0,0,0,0,0,0,4096);
    (0,0,0,0,0,0,0,0,8192);
    (0,0,0,0,0,0,0,0,16384);
    (0,0,0,0,0,0,0,0,32768);
    (0,0,0,0,0,0,0,0,65536);
    (0,0,0,0,0,0,0,0,131072);
    (0,0,0,0,0,0,0,0,262144);
    (0,0,0,0,0,0,0,0,524288);
    (0,0,0,0,0,0,0,0,1048576);
    (0,0,0,0,0,0,0,0,2097152);
    (0,0,0,0,0,0,0,0,4194304);
    (0,0,0,0,0,0,0,0,8388608);
    (0,0,0,0,0,0,0,0,16777216);
    (0,0,0,0,0,0,0,0,33554432);
    (0,0,0,0,0,0,0,0,67108864);
    (0,0,0,0,0,0,0,0,134217728);
    (0,0,0,0,0,0,0,0,268435456);
    (0,0,0,0,0,0,0,0,536870912);
    (0,0,0,0,0,0,0,1,0);
    (0,0,0,0,0,0,0,2,0);
    (0,0,0,0,0,0,0,4,0);
    (0,0,0,0,0,0,0,8,0);
    (0,0,0,0,0,0,0,16,0);
    (0,0,0,0,0,0,0,32,0);
    (0,0,0,0,0,0,0,64,0);
    (0,0,0,0,0,0,0,128,0);
    (0,0,0,0,0,0,0,256,0);
    (0,0,0,0,0,0,0,512,0);
    (0,0,0,0,0,0,0,1024,0);
    (0,0,0,0,0,0,0,2048,0);
    (0,0,0,0,0,0,0,4096,0);
    (0,0,0,0,0,0,0,8192,0);
    (0,0,0,0,0,0,0,16384,0);
    (0,0,0,0,0,0,0,32768,0);
    (0,0,0,0,0,0,0,65536,0);
    (0,0,0,0,0,0,0,131072,0);
    (0,0,0,0,0,0,0,262144,0);
    (0,0,0,0,0,0,0,524288,0);
    (0,0,0,0,0,0,0,1048576,0);
    (0,0,0,0,0,0,0,2097152,0);
    (0,0,0,0,0,0,0,4194304,0);
    (0,0,0,0,0,0,0,8388608,0);
    (0,0,0,0,0,0,0,16777216,0);
    (0,0,0,0,0,0,0,33554432,0);
    (0,0,0,0,0,0,0,67108864,0);
    (0,0,0,0,0,0,0,134217728,0);
    (0,0,0,0,0,0,0,268435456,0);
    (0,0,0,0,0,0,0,536870912,0);
    (0,0,0,0,0,0,1,0,0);
    (0,0,0,0,0,0,2,0,0);
    (0,0,0,0,0,0,4,0,0);
    (0,0,0,0,0,0,8,0,0);
    (0,0,0,0,0,0,16,0,0);
    (0,0,0,0,0,0,32,0,0);
    (0,0,0,0,0,0,64,0,0);
    (0,0,0,0,0,0,128,0,0);
    (0,0,0,0,0,0,256,0,0);
    (0,0,0,0,0,0,512,0,0);
    (0,0,0,0,0,0,1024,0,0);
    (0,0,0,0,0,0,2048,0,0);
    (0,0,0,0,0,0,4096,0,0);
    (0,0,0,0,0,0,8192,0,0);
    (0,0,0,0,0,0,16384,0,0);
    (0,0,0,0,0,0,32768,0,0);
    (0,0,0,0,0,0,65536,0,0);
    (0,0,0,0,0,0,131072,0,0);
    (0,0,0,0,0,0,262144,0,0);
    (0,0,0,0,0,0,524288,0,0);
    (0,0,0,0,0,0,1048576,0,0);
    (0,0,0,0,0,0,2097152,0,0);
    (0,0,0,0,0,0,4194304,0,0);
    (0,0,0,0,0,0,8388608,0,0);
    (0,0,0,0,0,0,16777216,0,0);
    (0,0,0,0,0,0,33554432,0,0);
    (0,0,0,0,0,0,67108864,0,0);
    (0,0,0,0,0,0,134217728,0,0);
    (0,0,0,0,0,0,268435456,0,0);
    (0,0,0,0,0,0,536870912,0,0);
    (0,0,0,0,0,1,0,0,0);
    (0,0,0,0,0,2,0,0,0);
    (0,0,0,0,0,4,0,0,0);
    (0,0,0,0,0,8,0,0,0);
    (0,0,0,0,0,16,0,0,0);
    (0,0,0,0,0,32,0,0,0);
    (0,0,0,0,0,64,0,0,0);
    (0,0,0,0,0,128,0,0,0);
    (0,0,0,0,0,256,0,0,0);
    (0,0,0,0,0,512,0,0,0);
    (0,0,0,0,0,1024,0,0,0);
    (0,0,0,0,0,2048,0,0,0);
    (0,0,0,0,0,4096,0,0,0);
    (0,0,0,0,0,8192,0,0,0);
    (0,0,0,0,0,16384,0,0,0);
    (0,0,0,0,0,32768,0,0,0);
    (0,0,0,0,0,65536,0,0,0);
    (0,0,0,0,0,131072,0,0,0);
    (0,0,0,0,0,262144,0,0,0);
    (0,0,0,0,0,524288,0,0,0);
    (0,0,0,0,0,1048576,0,0,0);
    (0,0,0,0,0,2097152,0,0,0);
    (0,0,0,0,0,4194304,0,0,0);
    (0,0,0,0,0,8388608,0,0,0);
    (0,0,0,0,0,16777216,0,0,0);
    (0,0,0,0,0,33554432,0,0,0);
    (0,0,0,0,0,67108864,0,0,0);
    (0,0,0,0,0,134217728,0,0,0);
    (0,0,0,0,0,268435456,0,0,0);
    (0,0,0,0,0,536870912,0,0,0);
    (0,0,0,0,1,0,0,0,0);
    (0,0,0,0,2,0,0,0,0);
    (0,0,0,0,4,0,0,0,0);
    (0,0,0,0,8,0,0,0,0);
    (0,0,0,0,16,0,0,0,0);
    (0,0,0,0,32,0,0,0,0);
    (0,0,0,0,64,0,0,0,0);
    (0,0,0,0,128,0,0,0,0);
    (0,0,0,0,256,0,0,0,0);
    (0,0,0,0,512,0,0,0,0);
    (0,0,0,0,1024,0,0,0,0);
    (0,0,0,0,2048,0,0,0,0);
    (0,0,0,0,4096,0,0,0,0);
    (0,0,0,0,8192,0,0,0,0);
    (0,0,0,0,16384,0,0,0,0);
    (0,0,0,0,32768,0,0,0,0);
    (0,0,0,0,65536,0,0,0,0);
    (0,0,0,0,131072,0,0,0,0);
    (0,0,0,0,262144,0,0,0,0);
    (0,0,0,0,524288,0,0,0,0);
    (0,0,0,0,1048576,0,0,0,0);
    (0,0,0,0,2097152,0,0,0,0);
    (0,0,0,0,4194304,0,0,0,0);
    (0,0,0,0,8388608,0,0,0,0);
    (0,0,0,0,16777216,0,0,0,0);
    (0,0,0,0,33554432,0,0,0,0);
    (0,0,0,0,67108864,0,0,0,0);
    (0,0,0,0,134217728,0,0,0,0);
    (0,0,0,0,268435456,0,0,0,0);
    (0,0,0,0,536870912,0,0,0,0);
    (0,0,0,1,0,0,0,0,0);
    (0,0,0,2,0,0,0,0,0);
    (0,0,0,4,0,0,0,0,0);
    (0,0,0,8,0,0,0,0,0);
    (0,0,0,16,0,0,0,0,0);
    (0,0,0,32,0,0,0,0,0);
    (0,0,0,64,0,0,0,0,0);
    (0,0,0,128,0,0,0,0,0);
    (0,0,0,256,0,0,0,0,0);
    (0,0,0,512,0,0,0,0,0);
    (0,0,0,1024,0,0,0,0,0);
    (0,0,0,2048,0,0,0,0,0);
    (0,0,0,4096,0,0,0,0,0);
    (0,0,0,8192,0,0,0,0,0);
    (0,0,0,16384,0,0,0,0,0);
    (0,0,0,32768,0,0,0,0,0);
    (0,0,0,65536,0,0,0,0,0);
    (0,0,0,131072,0,0,0,0,0);
    (0,0,0,262144,0,0,0,0,0);
    (0,0,0,524288,0,0,0,0,0);
    (0,0,0,1048576,0,0,0,0,0);
    (0,0,0,2097152,0,0,0,0,0);
    (0,0,0,4194304,0,0,0,0,0);
    (0,0,0,8388608,0,0,0,0,0);
    (0,0,0,16777216,0,0,0,0,0);
    (0,0,0,33554432,0,0,0,0,0);
    (0,0,0,67108864,0,0,0,0,0);
    (0,0,0,134217728,0,0,0,0,0);
    (0,0,0,268435456,0,0,0,0,0);
    (0,0,0,536870912,0,0,0,0,0);
    (0,0,1,0,0,0,0,0,0);
    (0,0,2,0,0,0,0,0,0);
    (0,0,4,0,0,0,0,0,0);
    (0,0,8,0,0,0,0,0,0);
    (0,0,16,0,0,0,0,0,0);
    (0,0,32,0,0,0,0,0,0);
    (0,0,64,0,0,0,0,0,0);
    (0,0,128,0,0,0,0,0,0);
    (0,0,256,0,0,0,0,0,0);
    (0,0,512,0,0,0,0,0,0);
    (0,0,1024,0,0,0,0,0,0);
    (0,0,2048,0,0,0,0,0,0);
    (0,0,4096,0,0,0,0,0,0);
    (0,0,8192,0,0,0,0,0,0);
    (0,0,16384,0,0,0,0,0,0);
    (0,0,32768,0,0,0,0,0,0);
    (0,0,65536,0,0,0,0,0,0);
    (0,0,131072,0,0,0,0,0,0);
    (0,0,262144,0,0,0,0,0,0);
    (0,0,524288,0,0,0,0,0,0);
    (0,0,1048576,0,0,0,0,0,0);
    (0,0,2097152,0,0,0,0,0,0);
    (0,0,4194304,0,0,0,0,0,0);
    (0,0,8388608,0,0,0,0,0,0);
    (0,0,16777216,0,0,0,0,0,0);
    (0,0,33554432,0,0,0,0,0,0);
    (0,0,67108864,0,0,0,0,0,0);
    (0,0,134217728,0,0,0,0,0,0);
    (0,0,268435456,0,0,0,0,0,0);
    (0,0,536870912,0,0,0,0,0,0);
    (0,1,0,0,0,0,0,0,0);
    (0,2,0,0,0,0,0,0,0);
    (0,4,0,0,0,0,0,0,0);
    (0,8,0,0,0,0,0,0,0);
    (0,16,0,0,0,0,0,0,0);
    (0,32,0,0,0,0,0,0,0);
    (0,64,0,0,0,0,0,0,0);
    (0,128,0,0,0,0,0,0,0);
    (0,256,0,0,0,0,0,0,0);
    (0,512,0,0,0,0,0,0,0);
    (0,1024,0,0,0,0,0,0,0);
    (0,2048,0,0,0,0,0,0,0);
    (0,4096,0,0,0,0,0,0,0);
    (0,8192,0,0,0,0,0,0,0);
    (0,16384,0,0,0,0,0,0,0);
    (0,32768,0,0,0,0,0,0,0);
    (0,65536,0,0,0,0,0,0,0);
    (0,131072,0,0,0,0,0,0,0);
    (0,262144,0,0,0,0,0,0,0);
    (0,524288,0,0,0,0,0,0,0);
    (0,1048576,0,0,0,0,0,0,0);
    (0,2097152,0,0,0,0,0,0,0);
    (0,4194304,0,0,0,0,0,0,0);
    (0,8388608,0,0,0,0,0,0,0);
    (0,16777216,0,0,0,0,0,0,0);
    (0,33554432,0,0,0,0,0,0,0);
    (0,67108864,0,0,0,0,0,0,0);
    (0,134217728,0,0,0,0,0,0,0);
    (0,268435456,0,0,0,0,0,0,0);
    (0,536870912,0,0,0,0,0,0,0);
    (1,0,0,0,0,0,0,0,0);
    (2,0,0,0,0,0,0,0,0);
    (4,0,0,0,0,0,0,0,0);
    (8,0,0,0,0,0,0,0,0);
    (16,0,0,0,0,0,0,0,0);
    (32,0,0,0,0,0,0,0,0);
    (64,0,0,0,0,0,0,0,0);
    (128,0,0,0,0,0,0,0,0);
    (256,0,0,0,0,0,0,0,0);
    (512,0,0,0,0,0,0,0,0);
    (1024,0,0,0,0,0,0,0,0);
    (2048,0,0,0,0,0,0,0,0);
    (4096,0,0,0,0,0,0,0,0);
    (8192,0,0,0,0,0,0,0,0);
    (16384,0,0,0,0,0,0,0,0);
    (32768,0,0,0,0,0,0,0,0);
    (0,0,0,0,0,0,0,4,977);
    (0,0,0,0,0,0,0,8,1954);
    (0,0,0,0,0,0,0,16,3908);
    (0,0,0,0,0,0,0,32,7816);
    (0,0,0,0,0,0,0,64,15632);
    (0,0,0,0,0,0,0,128,31264);
    (0,0,0,0,0,0,0,256,62528);
    (0,0,0,0,0,0,0,512,125056);
    (0,0,0,0,0,0,0,1024,250112);
    (0,0,0,0,0,0,0,2048,500224);
    (0,0,0,0,0,0,0,4096,1000448);
    (0,0,0,0,0,0,0,8192,2000896);
    (0,0,0,0,0,0,0,16384,4001792);
    (0,0,0,0,0,0,0,32768,8003584);
    (0,0,0,0,0,0,0,65536,16007168);
    (0,0,0,0,0,0,0,131072,32014336);
    (0,0,0,0,0,0,0,262144,64028672);
    (0,0,0,0,0,0,0,524288,128057344);
    (0,0,0,0,0,0,0,1048576,256114688);
    (0,0,0,0,0,0,0,2097152,512229376);
    (0,0,0,0,0,0,0,4194304,1024458752);
    (0,0,0,0,0,0,0,8388609,975175680);
    (0,0,0,0,0,0,0,16777219,876609536);
    (0,0,0,0,0,0,0,33554439,679477248);
    (0,0,0,0,0,0,0,67108879,285212672);
    (0,0,0,0,0,0,0,134217758,570425344);
    (0,0,0,0,0,0,0,268435517,67108864);
    (0,0,0,0,0,0,0,536871034,134217728);
    (0,0,0,0,0,0,1,244,268435456);
    (0,0,0,0,0,0,2,488,536870912);
    (0,0,0,0,0,0,4,977,0);
    (0,0,0,0,0,0,8,1954,0);
    (0,0,0,0,0,0,16,3908,0);
    (0,0,0,0,0,0,32,7816,0);
    (0,0,0,0,0,0,64,15632,0);
    (0,0,0,0,0,0,128,31264,0);
    (0,0,0,0,0,0,256,62528,0);
    (0,0,0,0,0,0,512,125056,0);
    (0,0,0,0,0,0,1024,250112,0);
    (0,0,0,0,0,0,2048,500224,0);
    (0,0,0,0,0,0,4096,1000448,0);
    (0,0,0,0,0,0,8192,2000896,0);
    (0,0,0,0,0,0,16384,4001792,0);
    (0,0,0,0,0,0,32768,8003584,0);
    (0,0,0,0,0,0,65536,16007168,0);
    (0,0,0,0,0,0,131072,32014336,0);
    (0,0,0,0,0,0,262144,64028672,0);
    (0,0,0,0,0,0,524288,128057344,0);
    (0,0,0,0,0,0,1048576,256114688,0);
    (0,0,0,0,0,0,2097152,512229376,0);
    (0,0,0,0,0,0,4194304,1024458752,0);
    (0,0,0,0,0,0,8388609,975175680,0);
    (0,0,0,0,0,0,16777219,876609536,0);
    (0,0,0,0,0,0,33554439,679477248,0);
    (0,0,0,0,0,0,67108879,285212672,0);
    (0,0,0,0,0,0,134217758,570425344,0);
    (0,0,0,0,0,0,268435517,67108864,0);
    (0,0,0,0,0,0,536871034,134217728,0);
    (0,0,0,0,0,1,244,268435456,0);
    (0,0,0,0,0,2,488,536870912,0);
    (0,0,0,0,0,4,977,0,0);
    (0,0,0,0,0,8,1954,0,0);
    (0,0,0,0,0,16,3908,0,0);
    (0,0,0,0,0,32,7816,0,0);
    (0,0,0,0,0,64,15632,0,0);
    (0,0,0,0,0,128,31264,0,0);
    (0,0,0,0,0,256,62528,0,0);
    (0,0,0,0,0,512,125056,0,0);
    (0,0,0,0,0,1024,250112,0,0);
    (0,0,0,0,0,2048,500224,0,0);
    (0,0,0,0,0,4096,1000448,0,0);
    (0,0,0,0,0,8192,2000896,0,0);
    (0,0,0,0,0,16384,4001792,0,0);
    (0,0,0,0,0,32768,8003584,0,0);
    (0,0,0,0,0,65536,16007168,0,0);
    (0,0,0,0,0,131072,32014336,0,0);
    (0,0,0,0,0,262144,64028672,0,0);
    (0,0,0,0,0,524288,128057344,0,0);
    (0,0,0,0,0,1048576,256114688,0,0);
    (0,0,0,0,0,2097152,512229376,0,0);
    (0,0,0,0,0,4194304,1024458752,0,0);
    (0,0,0,0,0,8388609,975175680,0,0);
    (0,0,0,0,0,16777219,876609536,0,0);
    (0,0,0,0,0,33554439,679477248,0,0);
    (0,0,0,0,0,67108879,285212672,0,0);
    (0,0,0,0,0,134217758,570425344,0,0);
    (0,0,0,0,0,268435517,67108864,0,0);
    (0,0,0,0,0,536871034,134217728,0,0);
    (0,0,0,0,1,244,268435456,0,0);
    (0,0,0,0,2,488,536870912,0,0);
    (0,0,0,0,4,977,0,0,0);
    (0,0,0,0,8,1954,0,0,0);
    (0,0,0,0,16,3908,0,0,0);
    (0,0,0,0,32,7816,0,0,0);
    (0,0,0,0,64,15632,0,0,0);
    (0,0,0,0,128,31264,0,0,0);
    (0,0,0,0,256,62528,0,0,0);
    (0,0,0,0,512,125056,0,0,0);
    (0,0,0,0,1024,250112,0,0,0);
    (0,0,0,0,2048,500224,0,0,0);
    (0,0,0,0,4096,1000448,0,0,0);
    (0,0,0,0,8192,2000896,0,0,0);
    (0,0,0,0,16384,4001792,0,0,0);
    (0,0,0,0,32768,8003584,0,0,0);
    (0,0,0,0,65536,16007168,0,0,0);
    (0,0,0,0,131072,32014336,0,0,0);
    (0,0,0,0,262144,64028672,0,0,0);
    (0,0,0,0,524288,128057344,0,0,0);
    (0,0,0,0,1048576,256114688,0,0,0);
    (0,0,0,0,2097152,512229376,0,0,0);
    (0,0,0,0,4194304,1024458752,0,0,0);
    (0,0,0,0,8388609,975175680,0,0,0);
    (0,0,0,0,16777219,876609536,0,0,0);
    (0,0,0,0,33554439,679477248,0,0,0);
    (0,0,0,0,67108879,285212672,0,0,0);
    (0,0,0,0,134217758,570425344,0,0,0);
    (0,0,0,0,268435517,67108864,0,0,0);
    (0,0,0,0,536871034,134217728,0,0,0);
    (0,0,0,1,244,268435456,0,0,0);
    (0,0,0,2,488,536870912,0,0,0);
    (0,0,0,4,977,0,0,0,0);
    (0,0,0,8,1954,0,0,0,0);
    (0,0,0,16,3908,0,0,0,0);
    (0,0,0,32,7816,0,0,0,0);
    (0,0,0,64,15632,0,0,0,0);
    (0,0,0,128,31264,0,0,0,0);
    (0,0,0,256,62528,0,0,0,0);
    (0,0,0,512,125056,0,0,0,0);
    (0,0,0,1024,250112,0,0,0,0);
    (0,0,0,2048,500224,0,0,0,0);
    (0,0,0,4096,1000448,0,0,0,0);
    (0,0,0,8192,2000896,0,0,0,0);
    (0,0,0,16384,4001792,0,0,0,0);
    (0,0,0,32768,8003584,0,0,0,0);
    (0,0,0,65536,16007168,0,0,0,0);
    (0,0,0,131072,32014336,0,0,0,0);
    (0,0,0,262144,64028672,0,0,0,0);
    (0,0,0,524288,128057344,0,0,0,0);
    (0,0,0,1048576,256114688,0,0,0,0);
    (0,0,0,2097152,512229376,0,0,0,0);
    (0,0,0,4194304,1024458752,0,0,0,0);
    (0,0,0,8388609,975175680,0,0,0,0);
    (0,0,0,16777219,876609536,0,0,0,0);
    (0,0,0,33554439,679477248,0,0,0,0);
    (0,0,0,67108879,285212672,0,0,0,0);
    (0,0,0,134217758,570425344,0,0,0,0);
    (0,0,0,268435517,67108864,0,0,0,0);
    (0,0,0,536871034,134217728,0,0,0,0);
    (0,0,1,244,268435456,0,0,0,0);
    (0,0,2,488,536870912,0,0,0,0);
    (0,0,4,977,0,0,0,0,0);
    (0,0,8,1954,0,0,0,0,0);
    (0,0,16,3908,0,0,0,0,0);
    (0,0,32,7816,0,0,0,0,0);
    (0,0,64,15632,0,0,0,0,0);
    (0,0,128,31264,0,0,0,0,0);
    (0,0,256,62528,0,0,0,0,0);
    (0,0,512,125056,0,0,0,0,0);
    (0,0,1024,250112,0,0,0,0,0);
    (0,0,2048,500224,0,0,0,0,0);
    (0,0,4096,1000448,0,0,0,0,0);
    (0,0,8192,2000896,0,0,0,0,0);
    (0,0,16384,4001792,0,0,0,0,0);
    (0,0,32768,8003584,0,0,0,0,0);
    (0,0,65536,16007168,0,0,0,0,0);
    (0,0,131072,32014336,0,0,0,0,0);
    (0,0,262144,64028672,0,0,0,0,0);
    (0,0,524288,128057344,0,0,0,0,0);
    (0,0,1048576,256114688,0,0,0,0,0);
    (0,0,2097152,512229376,0,0,0,0,0);
    (0,0,4194304,1024458752,0,0,0,0,0);
    (0,0,8388609,975175680,0,0,0,0,0);
    (0,0,16777219,876609536,0,0,0,0,0);
    (0,0,33554439,679477248,0,0,0,0,0);
    (0,0,67108879,285212672,0,0,0,0,0);
    (0,0,134217758,570425344,0,0,0,0,0);
    (0,0,268435517,67108864,0,0,0,0,0);
    (0,0,536871034,134217728,0,0,0,0,0);
    (0,1,244,268435456,0,0,0,0,0);
    (0,2,488,536870912,0,0,0,0,0);
    (0,4,977,0,0,0,0,0,0);
    (0,8,1954,0,0,0,0,0,0);
    (0,16,3908,0,0,0,0,0,0);
    (0,32,7816,0,0,0,0,0,0);
    (0,64,15632,0,0,0,0,0,0);
    (0,128,31264,0,0,0,0,0,0);
    (0,256,62528,0,0,0,0,0,0);
    (0,512,125056,0,0,0,0,0,0);
    (0,1024,250112,0,0,0,0,0,0);
    (0,2048,500224,0,0,0,0,0,0);
    (0,4096,1000448,0,0,0,0,0,0);
    (0,8192,2000896,0,0,0,0,0,0);
    (0,16384,4001792,0,0,0,0,0,0);
    (0,32768,8003584,0,0,0,0,0,0);
    (0,65536,16007168,0,0,0,0,0,0);
    (0,131072,32014336,0,0,0,0,0,0);
    (0,262144,64028672,0,0,0,0,0,0);
    (0,524288,128057344,0,0,0,0,0,0);
    (0,1048576,256114688,0,0,0,0,0,0);
    (0,2097152,512229376,0,0,0,0,0,0);
    (0,4194304,1024458752,0,0,0,0,0,0);
    (0,8388609,975175680,0,0,0,0,0,0);
    (0,16777219,876609536,0,0,0,0,0,0);
    (0,33554439,679477248,0,0,0,0,0,0);
    (0,67108879,285212672,0,0,0,0,0,0);
    (0,134217758,570425344,0,0,0,0,0,0);
    (0,268435517,67108864,0,0,0,0,0,0);
    (0,536871034,134217728,0,0,0,0,0,0);
    (1,244,268435456,0,0,0,0,0,0);
    (2,488,536870912,0,0,0,0,0,0);
    (4,977,0,0,0,0,0,0,0);
    (8,1954,0,0,0,0,0,0,0);
    (16,3908,0,0,0,0,0,0,0);
    (32,7816,0,0,0,0,0,0,0);
    (64,15632,0,0,0,0,0,0,0);
    (128,31264,0,0,0,0,0,0,0);
    (256,62528,0,0,0,0,0,0,0);
    (512,125056,0,0,0,0,0,0,0);
    (1024,250112,0,0,0,0,0,0,0);
    (2048,500224,0,0,0,0,0,0,0);
    (4096,1000448,0,0,0,0,0,0,0);
    (8192,2000896,0,0,0,0,0,0,0);
    (16384,4001792,0,0,0,0,0,0,0);
    (32768,8003584,0,0,0,0,0,0,0);
    (0,16007168,0,0,0,0,0,4,977);
    (0,32014336,0,0,0,0,0,8,1954);
    (0,64028672,0,0,0,0,0,16,3908);
    (0,128057344,0,0,0,0,0,32,7816);
    (0,256114688,0,0,0,0,0,64,15632);
    (0,512229376,0,0,0,0,0,128,31264);
    (0,1024458752,0,0,0,0,0,256,62528);
    (1,975175680,0,0,0,0,0,512,125056);
    (3,876609536,0,0,0,0,0,1024,250112);
    (7,679477248,0,0,0,0,0,2048,500224);
    (15,285212672,0,0,0,0,0,4096,1000448);
    (30,570425344,0,0,0,0,0,8192,2000896);
    (61,67108864,0,0,0,0,0,16384,4001792);
    (122,134217728,0,0,0,0,0,32768,8003584);
    (244,268435456,0,0,0,0,0,65536,16007168);
    (488,536870912,0,0,0,0,0,131072,32014336);
    (977,0,0,0,0,0,0,262144,64028672);
    (1954,0,0,0,0,0,0,524288,128057344);
    (3908,0,0,0,0,0,0,1048576,256114688);
    (7816,0,0,0,0,0,0,2097152,512229376);
    (15632,0,0,0,0,0,0,4194304,1024458752);
    (31264,0,0,0,0,0,0,8388609,975175680);
    (62528,0,0,0,0,0,0,16777219,876609536);
    (59520,0,0,0,0,0,0,33554443,679478225);
    (53504,0,0,0,0,0,0,67108891,285215603);
    (41472,0,0,0,0,0,0,134217786,570432183);
    (17408,0,0,0,0,0,0,268435577,67123519);
    (34816,0,0,0,0,0,0,536871154,134247038);
    (4096,0,0,0,0,0,1,488,268495053);
    (8192,0,0,0,0,0,2,976,536990106);
    (16384,0,0,0,0,0,4,1953,238388);
    (32768,0,0,0,0,0,8,3906,476776)
  |]

(* set reg2 to double of reg1 *)
let mul2_reg_intp reg1 reg2 =
  for i = 0 to 8 do reg2.(i) <- reg1.(i) done;
  add_reg_reg_intp reg1 reg2

(* set reg2 to triple of reg1 *)
let mul3_reg_intp reg1 reg2 =
  for i = 0 to 8 do reg2.(i) <- reg1.(i) done;
  add_reg_reg_intp reg1 reg2;
  add_reg_reg_intp reg1 reg2

(* multiply reg1 and x with result in reg2 *)
let mul_reg_intp reg1 x reg2 =
  for i = 0 to 8 do reg2.(i) <- 0 done;
  for j = 0 to 255 do boolarray.(j) <- intp_bit x j done;
  for i = 0 to 255 do
    if reg_bit reg1 i then
      for j = 0 to 255 do
	if boolarray.(j) then
	  add_reg_intp reg2 (intpb.(i + j))
      done
  done

let mul_reg_reg_intp reg1 reg2 =
  for j = 0 to 255 do boolarray.(j) <- reg_bit reg2 j done;
  for i = 0 to 8 do reg2.(i) <- 0 done;
  for i = 0 to 255 do
    if reg_bit reg1 i then
      for j = 0 to 255 do
	if boolarray.(j) then
	  add_reg_intp reg2 intpb.(i + j)
      done
  done

let square_reg_intp reg1 reg2 =
  for i = 0 to 8 do reg2.(i) <- 0 done;
  for j = 0 to 255 do boolarray.(j) <- reg_bit reg1 j done;
  for i = 0 to 255 do
    if boolarray.(i) then
      for j = 0 to 255 do
	if boolarray.(j) then
	  add_reg_intp reg2 intpb.(i + j)
      done
  done

let intp_lisp (x8,x7,x6,x5,x4,x3,x2,x1,x0) =
  Printf.sprintf "(+ %d (* %d (ash 1 30)) (* %d (ash 1 60)) (* %d (ash 1 90)) (* %d (ash 1 120)) (* %d (ash 1 150)) (* %d (ash 1 180)) (* %d (ash 1 210)) (* %d (ash 1 240)))" x0 x1 x2 x3 x4 x5 x6 x7 x8;;

let argc = Array.length Sys.argv;;
if argc < 3 then (Printf.printf "expected [add|mul|mul2|mul3|square] <filename>\n"; exit 1);;
let n = if argc > 3 then (try int_of_string (Sys.argv.(argc-3)) with _ -> 1) else 1;;
let op = Sys.argv.(argc-2);;
let fn = Sys.argv.(argc-1);;
if not (Sys.file_exists fn) then (Printf.printf "expected [add|mul|mul2|mul3|square] <filename>\n"; exit 1);;
for i = 1 to n do
if op = "add" then
  begin
    let f = open_in fn in
    try
      let cnt = ref 1 in
      while true do
        for i = 0 to 8 do
          let l = input_line f in
          rega.(i) <- int_of_string l
        done;
        for i = 0 to 8 do
          let l = input_line f in
          regb.(i) <- int_of_string l
        done;
        for i = 0 to 8 do
          let l = input_line f in
          regc.(i) <- int_of_string l
        done;
        add_reg_reg_intp rega regb;
        for i = 0 to 8 do
          if not (regb.(i) = regc.(i)) then (Printf.printf "Example %d failed.\n" !cnt; raise Exit)
        done;
        incr cnt
      done
    with _ -> close_in f
  end
else if op = "mul" then
  begin
    let f = open_in fn in
    try
      let cnt = ref 1 in
      while true do
        for i = 0 to 8 do
          let l = input_line f in
          rega.(i) <- int_of_string l
        done;
        for i = 0 to 8 do
          let l = input_line f in
          regb.(i) <- int_of_string l
        done;
        for i = 0 to 8 do
          let l = input_line f in
          regc.(i) <- int_of_string l
        done;
        mul_reg_reg_intp rega regb;
        for i = 0 to 8 do
          if not (regb.(i) = regc.(i)) then (Printf.printf "Example %d failed.\n" !cnt; raise Exit)
        done;
        incr cnt
      done
    with _ -> close_in f
  end
else if op = "mul2" then
  begin
    let f = open_in fn in
    try
      let cnt = ref 1 in
      while true do
        for i = 0 to 8 do
          let l = input_line f in
          rega.(i) <- int_of_string l
        done;
        for i = 0 to 8 do
          let l = input_line f in
          regc.(i) <- int_of_string l
        done;
        mul2_reg_intp rega regb;
        for i = 0 to 8 do
          if not (regb.(i) = regc.(i)) then (Printf.printf "Example %d failed.\n" !cnt; raise Exit)
        done;
        incr cnt
      done
    with _ -> close_in f
  end
else if op = "mul3" then
  begin
    let f = open_in fn in
    try
      let cnt = ref 1 in
      while true do
        for i = 0 to 8 do
          let l = input_line f in
          rega.(i) <- int_of_string l
        done;
        for i = 0 to 8 do
          let l = input_line f in
          regc.(i) <- int_of_string l
        done;
        mul3_reg_intp rega regb;
        for i = 0 to 8 do
          if not (regb.(i) = regc.(i)) then (Printf.printf "Example %d failed.\n" !cnt; raise Exit)
        done;
        incr cnt
      done
    with _ -> close_in f
  end
else if op = "square" then
  begin
    let f = open_in fn in
    try
      let cnt = ref 1 in
      while true do
        for i = 0 to 8 do
          let l = input_line f in
          rega.(i) <- int_of_string l
        done;
        for i = 0 to 8 do
          let l = input_line f in
          regc.(i) <- int_of_string l
        done;
        square_reg_intp rega regb;
        for i = 0 to 8 do
          if not (regb.(i) = regc.(i)) then (Printf.printf "Example %d failed.\n" !cnt; raise Exit)
        done;
        incr cnt
      done
    with _ -> close_in f
  end
else
  (Printf.printf "expected [add|mul|mul2|mul3|square] <filename>\n"; exit 1)
done;;
