(* base 2^62 *)
(* all mod p where p = 115792089237316195423570985008687907853269984665640564039457584007908834671663 *)
type intp = int * int * int * int * int

let zero_intp = (0,0,0,0,0)
let unit_intp = (0,0,0,0,1)
let two_intp = (0,0,0,0,2)
let p = (255,4611686018427387903,4611686018427387903,4611686018427387903,4611686014132419631)
let pa = [| 255;4611686018427387903;4611686018427387903;4611686018427387903;4611686014132419631 |]
let pminus2_intp = (255,4611686018427387903,4611686018427387903,4611686018427387903,4611686014132419629)

let rega : int array = Array.make 5 0
let regb : int array = Array.make 5 0
let regc : int array = Array.make 5 0
let regd : int array = Array.make 5 0
let rege : int array = Array.make 5 0
let regf : int array = Array.make 5 0
let regg : int array = Array.make 5 0
let regh : int array = Array.make 5 0

let lt_reg_p reg =
  if reg.(0) < 255 then
    true
  else if reg.(0) > 255 then
    false
  else
    if reg.(1) < 4611686018427387903 then
      true
    else if reg.(1) > 4611686018427387903 then
      false
    else
      if reg.(2) < 4611686018427387903 then
	true
      else if reg.(2) > 4611686018427387903 then
	false
      else
	if reg.(3) < 4611686018427387903 then
	  true
	else if reg.(3) > 4611686018427387903 then
	  false
	else
	  reg.(4) < 4611686014132419631

let sub_reg_p reg =
  let b = ref false in
  for i = 4 downto 0 do
    if !b then
      reg.(i) <- (reg.(i) - pa.(i)) - 1
    else
      reg.(i) <- reg.(i) - pa.(i);
    if reg.(i) < 0 then
      begin
	reg.(i) <- reg.(i) land 0x3fffffffffffffff;
	b := true
      end
    else
      b := false
  done

(** assume reg is in [0,2p) and change to be in [0,p) **)
let ensure_reg_modp reg = if not (lt_reg_p reg) then sub_reg_p reg

let reg_to_intp reg = (reg.(0),reg.(1),reg.(2),reg.(3),reg.(4))
let set_reg_intp reg (x4,x3,x2,x1,x0) =
  reg.(0) <- x4;
  reg.(1) <- x3;
  reg.(2) <- x2;
  reg.(3) <- x1;
  reg.(4) <- x0

let copy_reg_reg_intp reg1 reg2 =
  reg2.(0) <- reg1.(0);
  reg2.(1) <- reg1.(1);
  reg2.(2) <- reg1.(2);
  reg2.(3) <- reg1.(3);
  reg2.(4) <- reg1.(4)

let intp_i_r (x4,x3,x2,x1,x0) i =
  if i = 0 then x4
  else if i = 1 then x3
  else if i = 2 then x2
  else if i = 3 then x1
  else x0
      
(* add x to register with result in reg *)
let add_reg_intp reg x =
  let c = ref 0 in
  for i = 4 downto 0 do
    reg.(i) <- reg.(i) + intp_i_r x i + !c;
    if reg.(i) land 0x4000000000000000 = 0 then
      c := 0
    else
      begin
	reg.(i) <- reg.(i) land 0x3fffffffffffffff;
	c := 1
      end
  done;
  ensure_reg_modp reg

(* add two registers with result in reg2 *)
let add_reg_reg_intp reg1 reg2 =
  let c = ref 0 in
  for i = 4 downto 0 do
    reg2.(i) <- reg2.(i) + reg1.(i) + !c;
    if reg2.(i) land 0x4000000000000000 = 0 then
      c := 0
    else
      begin
	reg2.(i) <- reg2.(i) land 0x3fffffffffffffff;
	c := 1
      end
  done;
  ensure_reg_modp reg2

(* set reg to p - reg *)
let minus_reg_intp reg =
  reg.(0) <- pa.(0) - reg.(0);
  for i = 1 to 4 do
    reg.(i) <- pa.(i) - reg.(i);
    if reg.(i) < 0 then
      begin
	reg.(i-1) <- reg.(i-1) - 1;
	reg.(i) <- reg.(i) land 0x3fffffffffffffff
      end
  done

let reg_bit reg i =
  if i < 62 then
    (reg.(4) lsr i) land 1 = 1
  else if i < 124 then
    (reg.(3) lsr (i-62)) land 1 = 1
  else if i < 186 then
    (reg.(2) lsr (i-124)) land 1 = 1
  else if i < 248 then
    (reg.(1) lsr (i-186)) land 1 = 1
  else
    (reg.(0) lsr (i-248)) land 1 = 1

let intp_bit (x4,x3,x2,x1,x0) i =
  if i < 62 then
    (x0 lsr i) land 1 = 1
  else if i < 124 then
    (x1 lsr (i-62)) land 1 = 1
  else if i < 186 then
    (x2 lsr (i-124)) land 1 = 1
  else if i < 248 then
    (x3 lsr (i-186)) land 1 = 1
  else
    (x4 lsr (i-248)) land 1 = 1

let boolarray : bool array = Array.make 256 false
let intpb : intp array =
  [|
    (0,0,0,0,1);
    (0,0,0,0,2);
    (0,0,0,0,4);
    (0,0,0,0,8);
    (0,0,0,0,16);
    (0,0,0,0,32);
    (0,0,0,0,64);
    (0,0,0,0,128);
    (0,0,0,0,256);
    (0,0,0,0,512);
    (0,0,0,0,1024);
    (0,0,0,0,2048);
    (0,0,0,0,4096);
    (0,0,0,0,8192);
    (0,0,0,0,16384);
    (0,0,0,0,32768);
    (0,0,0,0,65536);
    (0,0,0,0,131072);
    (0,0,0,0,262144);
    (0,0,0,0,524288);
    (0,0,0,0,1048576);
    (0,0,0,0,2097152);
    (0,0,0,0,4194304);
    (0,0,0,0,8388608);
    (0,0,0,0,16777216);
    (0,0,0,0,33554432);
    (0,0,0,0,67108864);
    (0,0,0,0,134217728);
    (0,0,0,0,268435456);
    (0,0,0,0,536870912);
    (0,0,0,0,1073741824);
    (0,0,0,0,2147483648);
    (0,0,0,0,4294967296);
    (0,0,0,0,8589934592);
    (0,0,0,0,17179869184);
    (0,0,0,0,34359738368);
    (0,0,0,0,68719476736);
    (0,0,0,0,137438953472);
    (0,0,0,0,274877906944);
    (0,0,0,0,549755813888);
    (0,0,0,0,1099511627776);
    (0,0,0,0,2199023255552);
    (0,0,0,0,4398046511104);
    (0,0,0,0,8796093022208);
    (0,0,0,0,17592186044416);
    (0,0,0,0,35184372088832);
    (0,0,0,0,70368744177664);
    (0,0,0,0,140737488355328);
    (0,0,0,0,281474976710656);
    (0,0,0,0,562949953421312);
    (0,0,0,0,1125899906842624);
    (0,0,0,0,2251799813685248);
    (0,0,0,0,4503599627370496);
    (0,0,0,0,9007199254740992);
    (0,0,0,0,18014398509481984);
    (0,0,0,0,36028797018963968);
    (0,0,0,0,72057594037927936);
    (0,0,0,0,144115188075855872);
    (0,0,0,0,288230376151711744);
    (0,0,0,0,576460752303423488);
    (0,0,0,0,1152921504606846976);
    (0,0,0,0,2305843009213693952);
    (0,0,0,1,0);
    (0,0,0,2,0);
    (0,0,0,4,0);
    (0,0,0,8,0);
    (0,0,0,16,0);
    (0,0,0,32,0);
    (0,0,0,64,0);
    (0,0,0,128,0);
    (0,0,0,256,0);
    (0,0,0,512,0);
    (0,0,0,1024,0);
    (0,0,0,2048,0);
    (0,0,0,4096,0);
    (0,0,0,8192,0);
    (0,0,0,16384,0);
    (0,0,0,32768,0);
    (0,0,0,65536,0);
    (0,0,0,131072,0);
    (0,0,0,262144,0);
    (0,0,0,524288,0);
    (0,0,0,1048576,0);
    (0,0,0,2097152,0);
    (0,0,0,4194304,0);
    (0,0,0,8388608,0);
    (0,0,0,16777216,0);
    (0,0,0,33554432,0);
    (0,0,0,67108864,0);
    (0,0,0,134217728,0);
    (0,0,0,268435456,0);
    (0,0,0,536870912,0);
    (0,0,0,1073741824,0);
    (0,0,0,2147483648,0);
    (0,0,0,4294967296,0);
    (0,0,0,8589934592,0);
    (0,0,0,17179869184,0);
    (0,0,0,34359738368,0);
    (0,0,0,68719476736,0);
    (0,0,0,137438953472,0);
    (0,0,0,274877906944,0);
    (0,0,0,549755813888,0);
    (0,0,0,1099511627776,0);
    (0,0,0,2199023255552,0);
    (0,0,0,4398046511104,0);
    (0,0,0,8796093022208,0);
    (0,0,0,17592186044416,0);
    (0,0,0,35184372088832,0);
    (0,0,0,70368744177664,0);
    (0,0,0,140737488355328,0);
    (0,0,0,281474976710656,0);
    (0,0,0,562949953421312,0);
    (0,0,0,1125899906842624,0);
    (0,0,0,2251799813685248,0);
    (0,0,0,4503599627370496,0);
    (0,0,0,9007199254740992,0);
    (0,0,0,18014398509481984,0);
    (0,0,0,36028797018963968,0);
    (0,0,0,72057594037927936,0);
    (0,0,0,144115188075855872,0);
    (0,0,0,288230376151711744,0);
    (0,0,0,576460752303423488,0);
    (0,0,0,1152921504606846976,0);
    (0,0,0,2305843009213693952,0);
    (0,0,1,0,0);
    (0,0,2,0,0);
    (0,0,4,0,0);
    (0,0,8,0,0);
    (0,0,16,0,0);
    (0,0,32,0,0);
    (0,0,64,0,0);
    (0,0,128,0,0);
    (0,0,256,0,0);
    (0,0,512,0,0);
    (0,0,1024,0,0);
    (0,0,2048,0,0);
    (0,0,4096,0,0);
    (0,0,8192,0,0);
    (0,0,16384,0,0);
    (0,0,32768,0,0);
    (0,0,65536,0,0);
    (0,0,131072,0,0);
    (0,0,262144,0,0);
    (0,0,524288,0,0);
    (0,0,1048576,0,0);
    (0,0,2097152,0,0);
    (0,0,4194304,0,0);
    (0,0,8388608,0,0);
    (0,0,16777216,0,0);
    (0,0,33554432,0,0);
    (0,0,67108864,0,0);
    (0,0,134217728,0,0);
    (0,0,268435456,0,0);
    (0,0,536870912,0,0);
    (0,0,1073741824,0,0);
    (0,0,2147483648,0,0);
    (0,0,4294967296,0,0);
    (0,0,8589934592,0,0);
    (0,0,17179869184,0,0);
    (0,0,34359738368,0,0);
    (0,0,68719476736,0,0);
    (0,0,137438953472,0,0);
    (0,0,274877906944,0,0);
    (0,0,549755813888,0,0);
    (0,0,1099511627776,0,0);
    (0,0,2199023255552,0,0);
    (0,0,4398046511104,0,0);
    (0,0,8796093022208,0,0);
    (0,0,17592186044416,0,0);
    (0,0,35184372088832,0,0);
    (0,0,70368744177664,0,0);
    (0,0,140737488355328,0,0);
    (0,0,281474976710656,0,0);
    (0,0,562949953421312,0,0);
    (0,0,1125899906842624,0,0);
    (0,0,2251799813685248,0,0);
    (0,0,4503599627370496,0,0);
    (0,0,9007199254740992,0,0);
    (0,0,18014398509481984,0,0);
    (0,0,36028797018963968,0,0);
    (0,0,72057594037927936,0,0);
    (0,0,144115188075855872,0,0);
    (0,0,288230376151711744,0,0);
    (0,0,576460752303423488,0,0);
    (0,0,1152921504606846976,0,0);
    (0,0,2305843009213693952,0,0);
    (0,1,0,0,0);
    (0,2,0,0,0);
    (0,4,0,0,0);
    (0,8,0,0,0);
    (0,16,0,0,0);
    (0,32,0,0,0);
    (0,64,0,0,0);
    (0,128,0,0,0);
    (0,256,0,0,0);
    (0,512,0,0,0);
    (0,1024,0,0,0);
    (0,2048,0,0,0);
    (0,4096,0,0,0);
    (0,8192,0,0,0);
    (0,16384,0,0,0);
    (0,32768,0,0,0);
    (0,65536,0,0,0);
    (0,131072,0,0,0);
    (0,262144,0,0,0);
    (0,524288,0,0,0);
    (0,1048576,0,0,0);
    (0,2097152,0,0,0);
    (0,4194304,0,0,0);
    (0,8388608,0,0,0);
    (0,16777216,0,0,0);
    (0,33554432,0,0,0);
    (0,67108864,0,0,0);
    (0,134217728,0,0,0);
    (0,268435456,0,0,0);
    (0,536870912,0,0,0);
    (0,1073741824,0,0,0);
    (0,2147483648,0,0,0);
    (0,4294967296,0,0,0);
    (0,8589934592,0,0,0);
    (0,17179869184,0,0,0);
    (0,34359738368,0,0,0);
    (0,68719476736,0,0,0);
    (0,137438953472,0,0,0);
    (0,274877906944,0,0,0);
    (0,549755813888,0,0,0);
    (0,1099511627776,0,0,0);
    (0,2199023255552,0,0,0);
    (0,4398046511104,0,0,0);
    (0,8796093022208,0,0,0);
    (0,17592186044416,0,0,0);
    (0,35184372088832,0,0,0);
    (0,70368744177664,0,0,0);
    (0,140737488355328,0,0,0);
    (0,281474976710656,0,0,0);
    (0,562949953421312,0,0,0);
    (0,1125899906842624,0,0,0);
    (0,2251799813685248,0,0,0);
    (0,4503599627370496,0,0,0);
    (0,9007199254740992,0,0,0);
    (0,18014398509481984,0,0,0);
    (0,36028797018963968,0,0,0);
    (0,72057594037927936,0,0,0);
    (0,144115188075855872,0,0,0);
    (0,288230376151711744,0,0,0);
    (0,576460752303423488,0,0,0);
    (0,1152921504606846976,0,0,0);
    (0,2305843009213693952,0,0,0);
    (1,0,0,0,0);
    (2,0,0,0,0);
    (4,0,0,0,0);
    (8,0,0,0,0);
    (16,0,0,0,0);
    (32,0,0,0,0);
    (64,0,0,0,0);
    (128,0,0,0,0);
    (0,0,0,0,4294968273);
    (0,0,0,0,8589936546);
    (0,0,0,0,17179873092);
    (0,0,0,0,34359746184);
    (0,0,0,0,68719492368);
    (0,0,0,0,137438984736);
    (0,0,0,0,274877969472);
    (0,0,0,0,549755938944);
    (0,0,0,0,1099511877888);
    (0,0,0,0,2199023755776);
    (0,0,0,0,4398047511552);
    (0,0,0,0,8796095023104);
    (0,0,0,0,17592190046208);
    (0,0,0,0,35184380092416);
    (0,0,0,0,70368760184832);
    (0,0,0,0,140737520369664);
    (0,0,0,0,281475040739328);
    (0,0,0,0,562950081478656);
    (0,0,0,0,1125900162957312);
    (0,0,0,0,2251800325914624);
    (0,0,0,0,4503600651829248);
    (0,0,0,0,9007201303658496);
    (0,0,0,0,18014402607316992);
    (0,0,0,0,36028805214633984);
    (0,0,0,0,72057610429267968);
    (0,0,0,0,144115220858535936);
    (0,0,0,0,288230441717071872);
    (0,0,0,0,576460883434143744);
    (0,0,0,0,1152921766868287488);
    (0,0,0,0,2305843533736574976);
    (0,0,0,1,1049045762048);
    (0,0,0,2,2098091524096);
    (0,0,0,4,4196183048192);
    (0,0,0,8,8392366096384);
    (0,0,0,16,16784732192768);
    (0,0,0,32,33569464385536);
    (0,0,0,64,67138928771072);
    (0,0,0,128,134277857542144);
    (0,0,0,256,268555715084288);
    (0,0,0,512,537111430168576);
    (0,0,0,1024,1074222860337152);
    (0,0,0,2048,2148445720674304);
    (0,0,0,4096,4296891441348608);
    (0,0,0,8192,8593782882697216);
    (0,0,0,16384,17187565765394432);
    (0,0,0,32768,34375131530788864);
    (0,0,0,65536,68750263061577728);
    (0,0,0,131072,137500526123155456);
    (0,0,0,262144,275001052246310912);
    (0,0,0,524288,550002104492621824);
    (0,0,0,1048576,1100004208985243648);
    (0,0,0,2097152,2200008417970487296);
    (0,0,0,4194304,4400016835940974592);
    (0,0,0,8388609,4188347653454561280);
    (0,0,0,16777219,3765009288481734656);
    (0,0,0,33554439,2918332558536081408);
    (0,0,0,67108879,1224979098644774912);
    (0,0,0,134217758,2449958197289549824);
    (0,0,0,268435517,288230376151711744);
    (0,0,0,536871034,576460752303423488);
    (0,0,0,1073742068,1152921504606846976);
    (0,0,0,2147484136,2305843009213693952);
    (0,0,0,4294968273,0);
    (0,0,0,8589936546,0);
    (0,0,0,17179873092,0);
    (0,0,0,34359746184,0);
    (0,0,0,68719492368,0);
    (0,0,0,137438984736,0);
    (0,0,0,274877969472,0);
    (0,0,0,549755938944,0);
    (0,0,0,1099511877888,0);
    (0,0,0,2199023755776,0);
    (0,0,0,4398047511552,0);
    (0,0,0,8796095023104,0);
    (0,0,0,17592190046208,0);
    (0,0,0,35184380092416,0);
    (0,0,0,70368760184832,0);
    (0,0,0,140737520369664,0);
    (0,0,0,281475040739328,0);
    (0,0,0,562950081478656,0);
    (0,0,0,1125900162957312,0);
    (0,0,0,2251800325914624,0);
    (0,0,0,4503600651829248,0);
    (0,0,0,9007201303658496,0);
    (0,0,0,18014402607316992,0);
    (0,0,0,36028805214633984,0);
    (0,0,0,72057610429267968,0);
    (0,0,0,144115220858535936,0);
    (0,0,0,288230441717071872,0);
    (0,0,0,576460883434143744,0);
    (0,0,0,1152921766868287488,0);
    (0,0,0,2305843533736574976,0);
    (0,0,1,1049045762048,0);
    (0,0,2,2098091524096,0);
    (0,0,4,4196183048192,0);
    (0,0,8,8392366096384,0);
    (0,0,16,16784732192768,0);
    (0,0,32,33569464385536,0);
    (0,0,64,67138928771072,0);
    (0,0,128,134277857542144,0);
    (0,0,256,268555715084288,0);
    (0,0,512,537111430168576,0);
    (0,0,1024,1074222860337152,0);
    (0,0,2048,2148445720674304,0);
    (0,0,4096,4296891441348608,0);
    (0,0,8192,8593782882697216,0);
    (0,0,16384,17187565765394432,0);
    (0,0,32768,34375131530788864,0);
    (0,0,65536,68750263061577728,0);
    (0,0,131072,137500526123155456,0);
    (0,0,262144,275001052246310912,0);
    (0,0,524288,550002104492621824,0);
    (0,0,1048576,1100004208985243648,0);
    (0,0,2097152,2200008417970487296,0);
    (0,0,4194304,4400016835940974592,0);
    (0,0,8388609,4188347653454561280,0);
    (0,0,16777219,3765009288481734656,0);
    (0,0,33554439,2918332558536081408,0);
    (0,0,67108879,1224979098644774912,0);
    (0,0,134217758,2449958197289549824,0);
    (0,0,268435517,288230376151711744,0);
    (0,0,536871034,576460752303423488,0);
    (0,0,1073742068,1152921504606846976,0);
    (0,0,2147484136,2305843009213693952,0);
    (0,0,4294968273,0,0);
    (0,0,8589936546,0,0);
    (0,0,17179873092,0,0);
    (0,0,34359746184,0,0);
    (0,0,68719492368,0,0);
    (0,0,137438984736,0,0);
    (0,0,274877969472,0,0);
    (0,0,549755938944,0,0);
    (0,0,1099511877888,0,0);
    (0,0,2199023755776,0,0);
    (0,0,4398047511552,0,0);
    (0,0,8796095023104,0,0);
    (0,0,17592190046208,0,0);
    (0,0,35184380092416,0,0);
    (0,0,70368760184832,0,0);
    (0,0,140737520369664,0,0);
    (0,0,281475040739328,0,0);
    (0,0,562950081478656,0,0);
    (0,0,1125900162957312,0,0);
    (0,0,2251800325914624,0,0);
    (0,0,4503600651829248,0,0);
    (0,0,9007201303658496,0,0);
    (0,0,18014402607316992,0,0);
    (0,0,36028805214633984,0,0);
    (0,0,72057610429267968,0,0);
    (0,0,144115220858535936,0,0);
    (0,0,288230441717071872,0,0);
    (0,0,576460883434143744,0,0);
    (0,0,1152921766868287488,0,0);
    (0,0,2305843533736574976,0,0);
    (0,1,1049045762048,0,0);
    (0,2,2098091524096,0,0);
    (0,4,4196183048192,0,0);
    (0,8,8392366096384,0,0);
    (0,16,16784732192768,0,0);
    (0,32,33569464385536,0,0);
    (0,64,67138928771072,0,0);
    (0,128,134277857542144,0,0);
    (0,256,268555715084288,0,0);
    (0,512,537111430168576,0,0);
    (0,1024,1074222860337152,0,0);
    (0,2048,2148445720674304,0,0);
    (0,4096,4296891441348608,0,0);
    (0,8192,8593782882697216,0,0);
    (0,16384,17187565765394432,0,0);
    (0,32768,34375131530788864,0,0);
    (0,65536,68750263061577728,0,0);
    (0,131072,137500526123155456,0,0);
    (0,262144,275001052246310912,0,0);
    (0,524288,550002104492621824,0,0);
    (0,1048576,1100004208985243648,0,0);
    (0,2097152,2200008417970487296,0,0);
    (0,4194304,4400016835940974592,0,0);
    (0,8388609,4188347653454561280,0,0);
    (0,16777219,3765009288481734656,0,0);
    (0,33554439,2918332558536081408,0,0);
    (0,67108879,1224979098644774912,0,0);
    (0,134217758,2449958197289549824,0,0);
    (0,268435517,288230376151711744,0,0);
    (0,536871034,576460752303423488,0,0);
    (0,1073742068,1152921504606846976,0,0);
    (0,2147484136,2305843009213693952,0,0);
    (0,4294968273,0,0,0);
    (0,8589936546,0,0,0);
    (0,17179873092,0,0,0);
    (0,34359746184,0,0,0);
    (0,68719492368,0,0,0);
    (0,137438984736,0,0,0);
    (0,274877969472,0,0,0);
    (0,549755938944,0,0,0);
    (0,1099511877888,0,0,0);
    (0,2199023755776,0,0,0);
    (0,4398047511552,0,0,0);
    (0,8796095023104,0,0,0);
    (0,17592190046208,0,0,0);
    (0,35184380092416,0,0,0);
    (0,70368760184832,0,0,0);
    (0,140737520369664,0,0,0);
    (0,281475040739328,0,0,0);
    (0,562950081478656,0,0,0);
    (0,1125900162957312,0,0,0);
    (0,2251800325914624,0,0,0);
    (0,4503600651829248,0,0,0);
    (0,9007201303658496,0,0,0);
    (0,18014402607316992,0,0,0);
    (0,36028805214633984,0,0,0);
    (0,72057610429267968,0,0,0);
    (0,144115220858535936,0,0,0);
    (0,288230441717071872,0,0,0);
    (0,576460883434143744,0,0,0);
    (0,1152921766868287488,0,0,0);
    (0,2305843533736574976,0,0,0);
    (1,1049045762048,0,0,0);
    (2,2098091524096,0,0,0);
    (4,4196183048192,0,0,0);
    (8,8392366096384,0,0,0);
    (16,16784732192768,0,0,0);
    (32,33569464385536,0,0,0);
    (64,67138928771072,0,0,0);
    (128,134277857542144,0,0,0);
    (0,268555715084288,0,0,4294968273);
    (0,537111430168576,0,0,8589936546);
    (0,1074222860337152,0,0,17179873092);
    (0,2148445720674304,0,0,34359746184);
    (0,4296891441348608,0,0,68719492368);
    (0,8593782882697216,0,0,137438984736);
    (0,17187565765394432,0,0,274877969472);
    (0,34375131530788864,0,0,549755938944);
    (0,68750263061577728,0,0,1099511877888);
    (0,137500526123155456,0,0,2199023755776);
    (0,275001052246310912,0,0,4398047511552);
    (0,550002104492621824,0,0,8796095023104);
    (0,1100004208985243648,0,0,17592190046208);
    (0,2200008417970487296,0,0,35184380092416);
    (0,4400016835940974592,0,0,70368760184832);
    (1,4188347653454561280,0,0,140737520369664);
    (3,3765009288481734656,0,0,281475040739328);
    (7,2918332558536081408,0,0,562950081478656);
    (15,1224979098644774912,0,0,1125900162957312);
    (30,2449958197289549824,0,0,2251800325914624);
    (61,288230376151711744,0,0,4503600651829248);
    (122,576460752303423488,0,0,9007201303658496);
    (244,1152921504606846976,0,0,18014402607316992);
    (232,2305843009213693952,0,0,36028809509602257);
    (209,0,0,0,72057623314172787);
    (162,0,0,0,144115250923313847);
    (68,0,0,0,288230506141595967);
    (136,0,0,0,576461012283191934);
    (16,0,0,0,1152922028861352141);
    (32,0,0,0,2305844057722704282);
    (64,0,0,1,2097018020660);
  |]

(* set reg2 to double of reg1 *)
let mul2_reg_intp reg1 reg2 =
  for i = 0 to 4 do reg2.(i) <- reg1.(i) done;
  add_reg_reg_intp reg1 reg2

(* set reg2 to triple of reg1 *)
let mul3_reg_intp reg1 reg2 =
  for i = 0 to 4 do reg2.(i) <- reg1.(i) done;
  add_reg_reg_intp reg1 reg2;
  add_reg_reg_intp reg1 reg2

(* multiply reg1 and x with result in reg2 *)
let mul_reg_intp reg1 x reg2 =
  for i = 0 to 4 do reg2.(i) <- 0 done;
  for j = 0 to 255 do boolarray.(j) <- intp_bit x j done;
  for i = 0 to 255 do
    if reg_bit reg1 i then
      for j = 0 to 255 do
	if boolarray.(j) then
	  add_reg_intp reg2 (intpb.(i + j))
      done
  done

let mul_reg_reg_intp reg1 reg2 =
  for j = 0 to 255 do boolarray.(j) <- reg_bit reg2 j done;
  for i = 0 to 4 do reg2.(i) <- 0 done;
  for i = 0 to 255 do
    if reg_bit reg1 i then
      for j = 0 to 255 do
	if boolarray.(j) then
	  add_reg_intp reg2 intpb.(i + j)
      done
  done

let square_reg_intp reg1 reg2 =
  for i = 0 to 4 do reg2.(i) <- 0 done;
  for j = 0 to 255 do boolarray.(j) <- reg_bit reg1 j done;
  for i = 0 to 255 do
    if boolarray.(i) then
      for j = 0 to 255 do
	if boolarray.(j) then
	  add_reg_intp reg2 intpb.(i + j)
      done
  done

let argc = Array.length Sys.argv;;
if argc < 3 then (Printf.printf "expected [add|mul|mul2|mul3|square] <filename>\n"; exit 1);;
let n = if argc > 3 then (try int_of_string (Sys.argv.(argc-3)) with _ -> 1) else 1;;
let op = Sys.argv.(argc-2);;
let fn = Sys.argv.(argc-1);;
if not (Sys.file_exists fn) then (Printf.printf "expected [add|mul|mul2|mul3|square] <filename>\n"; exit 1);;
for i = 1 to n do
if op = "add" then
  begin
    let f = open_in fn in
    try
      let cnt = ref 1 in
      while true do
        for i = 0 to 4 do
          let l = input_line f in
          rega.(i) <- int_of_string l
        done;
        for i = 0 to 4 do
          let l = input_line f in
          regb.(i) <- int_of_string l
        done;
        for i = 0 to 4 do
          let l = input_line f in
          regc.(i) <- int_of_string l
        done;
        add_reg_reg_intp rega regb;
        for i = 0 to 4 do
          if not (regb.(i) = regc.(i)) then (Printf.printf "Example %d failed.\n" !cnt; raise Exit)
        done;
        incr cnt
      done
    with _ -> close_in f
  end
else if op = "mul" then
  begin
    let f = open_in fn in
    try
      let cnt = ref 1 in
      while true do
        for i = 0 to 4 do
          let l = input_line f in
          rega.(i) <- int_of_string l
        done;
        for i = 0 to 4 do
          let l = input_line f in
          regb.(i) <- int_of_string l
        done;
        for i = 0 to 4 do
          let l = input_line f in
          regc.(i) <- int_of_string l
        done;
        mul_reg_reg_intp rega regb;
        for i = 0 to 4 do
          if not (regb.(i) = regc.(i)) then (Printf.printf "Example %d failed.\n" !cnt; raise Exit)
        done;
        incr cnt
      done
    with _ -> close_in f
  end
else if op = "mul2" then
  begin
    let f = open_in fn in
    try
      let cnt = ref 1 in
      while true do
        for i = 0 to 4 do
          let l = input_line f in
          rega.(i) <- int_of_string l
        done;
        for i = 0 to 4 do
          let l = input_line f in
          regc.(i) <- int_of_string l
        done;
        mul2_reg_intp rega regb;
        for i = 0 to 4 do
          if not (regb.(i) = regc.(i)) then (Printf.printf "Example %d failed.\n" !cnt; raise Exit)
        done;
        incr cnt
      done
    with _ -> close_in f
  end
else if op = "mul3" then
  begin
    let f = open_in fn in
    try
      let cnt = ref 1 in
      while true do
        for i = 0 to 4 do
          let l = input_line f in
          rega.(i) <- int_of_string l
        done;
        for i = 0 to 4 do
          let l = input_line f in
          regc.(i) <- int_of_string l
        done;
        mul3_reg_intp rega regb;
        for i = 0 to 4 do
          if not (regb.(i) = regc.(i)) then (Printf.printf "Example %d failed.\n" !cnt; raise Exit)
        done;
        incr cnt
      done
    with _ -> close_in f
  end
else if op = "square" then
  begin
    let f = open_in fn in
    try
      let cnt = ref 1 in
      while true do
        for i = 0 to 4 do
          let l = input_line f in
          rega.(i) <- int_of_string l
        done;
        for i = 0 to 4 do
          let l = input_line f in
          regc.(i) <- int_of_string l
        done;
        square_reg_intp rega regb;
        for i = 0 to 4 do
          if not (regb.(i) = regc.(i)) then (Printf.printf "Example %d failed.\n" !cnt; raise Exit)
        done;
        incr cnt
      done
    with _ -> close_in f
  end
else
  (Printf.printf "expected [add|mul|mul2|mul3|square] <filename>\n"; exit 1)
done;;
